/*********************************************************************NVMH3****
Copyright NVIDIA Corporation 2003
TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
*AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY
LOSS) ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF
NVIDIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
******************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "READ_BMP.h"

#include <GL/gl.h>
#include "glut.h"
#include <Cg/cg.h>
#include <Cg/cgGL.h>

//Cg program, profiles...
static const char **compilerArgs;
static CGcontext context;
static CGprofile vertexProfile = CG_PROFILE_UNKNOWN;
static CGprofile fragmentProfile = CG_PROFILE_UNKNOWN;
static CGprogram vertexProg, vertexProg2, fragment_prog1, fragment_prog2, fragment_prog3, fragment_prog4, fragment_prog5;

//The camera system.
float camera_distance=45;
float camera_rotate_angle=0;
float camera_elevate_angle=45;
int screenWidth = 400;
int screenHeight = 400;
int mouseX, mouseY;
bool moveMode;

//The single light position.
float lightPosition[3]={0,6,6};

//Image textures.
GLuint texture_id;
BYTE * image_data;
int image_width;
int image_height;
CGparameter image_texture;

//OpenGL Callback Functions
static void initPrograms();
static void Handle_Display();
static void Handle_Keyboard(unsigned char key, int x, int y);
static void Handle_Reshape(int, int);
static void Handle_Motion(int, int);
static void Handle_Mouse(int, int, int, int);
//Cg Callback Function.
static void Handle_Cg_Error();


static const char ** parseArgs(int argc, const char **argv)
{
	// Huamin Wang: It's fine if you ignore these profile argments. The program loads the previously used profile argments in initPrograms.
	//profile types.
	//vp20, vp30, vp40 [NV_vertex_program 1.0, NV_vertex_program 2.0, and NV_vertex_program 3.0] 
	//fp30, fp40 [NV30 and NV40 OpenGL fragment programs] 
	//fp20 [NV_register_combiners and NV_Texture_shader) 
	//vs_1_1 for DirectX 8 and DirectX 9 
	//vs_2_0 and vs_2_x for DirectX 9 
	//ps_1_1, ps_1_2 and ps_1_3 for DirectX 8 and DirectX 9 
	//ps_2_0 and ps_2_x for DirectX 9 
	//arbvp1 [OpenGL ARB_vertex_program] 
	//arbfp1 [OpenGL ARB_fragment_program]
    argv++;
    while (--argc) {
        if (0 == strcmp(*argv, "-nv3x")) {
            vertexProfile = CG_PROFILE_VP30;
            fragmentProfile = CG_PROFILE_FP30;
        } else if (0 == strcmp(*argv, "-arb")) {
            vertexProfile = CG_PROFILE_ARBVP1;
            fragmentProfile = CG_PROFILE_ARBFP1;
        } else if (0 == strcmp(*argv, "-nv4x")) {
            vertexProfile = CG_PROFILE_VP30;
            fragmentProfile = CG_PROFILE_FP40;
        } else {
            // return pointer to first non-recognized option
            return argv;
        }
        argv++;
    }
    return argv;
}

void transformPoint(float p[3], float m[16], float ov[3])
{
    ov[0] = m[0]*p[0] + m[4]*p[1] + m[8]*p[2] + m[12];
    ov[1] = m[1]*p[0] + m[5]*p[1] + m[9]*p[2] + m[13];
    ov[2] = m[2]*p[0] + m[6]*p[1] + m[10]*p[2] + m[14];
    float w = m[3]*p[0] + m[7]*p[1] + m[11]*p[2] + m[15];
    if (w != 1.) { ov[0] /= w; ov[1] /= w; ov[2] /= w; }
}

///////////////////////////////////////////////////////////////////////////
// Main program; do basic GLUT and Cg setup, but leave most of the work to the display() function.

int main(int argc, char *argv[]) 
{
    compilerArgs = parseArgs(argc, (const char **)argv);
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
    glutInitWindowSize(screenWidth, screenHeight);
    glutCreateWindow("cgGL interfaces example");

    glutMouseFunc(Handle_Mouse);
    glutReshapeFunc(Handle_Reshape);
    glutKeyboardFunc(Handle_Keyboard);
    glutMotionFunc(Handle_Motion);
    glutDisplayFunc(Handle_Display);

	glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);

	BMP_Read("../duck.bmp", &image_data, image_width, image_height);
	glEnable (GL_TEXTURE_2D);
	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glGenTextures( 1, &texture_id );
	glBindTexture (GL_TEXTURE_2D,  texture_id);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGB, image_width, image_height,0, GL_RGB, GL_UNSIGNED_BYTE, image_data );

    initPrograms();    
    glutMainLoop();// and all the rest happens in the display function...
    return 0;
}

static void bindPrograms(CGprogram vert, CGprogram frag)
{
    CGparameter param;
    cgGLEnableProfile(vertexProfile);
    cgGLEnableProfile(fragmentProfile);
    cgGLBindProgram(vert);
    cgGLBindProgram(frag);
    param = cgGetNamedProgramParameter(vert, CG_PROGRAM, "ModelView");
    cgGLSetStateMatrixParameter(param, CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_IDENTITY);
    param = cgGetNamedProgramParameter(vert, CG_PROGRAM, "ModelViewIT");
    cgGLSetStateMatrixParameter(param, CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_INVERSE_TRANSPOSE);
    param = cgGetNamedProgramParameter(vert, CG_PROGRAM, "ModelViewProj");
    cgGLSetStateMatrixParameter(param, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);

    float modelViewMat[16], lightPosEye[3];
    glGetFloatv(GL_MODELVIEW_MATRIX, modelViewMat);
    transformPoint(lightPosition, modelViewMat, lightPosEye);
    param = cgGetNamedProgramParameter(vert, CG_PROGRAM, "lightPosition");
    cgSetParameter3fv(param, lightPosEye);
}
   
static void Handle_Display(void)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
	gluLookAt(0,45,0,0,0,0, 0,0,1);
	glRotatef(camera_elevate_angle, -1,0,0);
	glRotatef(camera_rotate_angle, 0,0,1);

    bindPrograms(vertexProg, fragment_prog5);  //the ground plane
	glNormal3f(0,0,1);
	glBegin(GL_QUADS);
	glVertex3d(-10.0, -10.0, 0.0);
	glVertex3d( 10.0, -10.0, 0.0);
	glVertex3d( 10.0,  10.0, 0.0);
	glVertex3d(-10.0,  10.0, 0.0);
	glEnd();

	bindPrograms(vertexProg2, fragment_prog4);	//board 4
	glNormal3f(0,1,0);
	glBegin(GL_QUADS);
	for(float i = 0; i <= 5.5; i += 0.5) {
		glVertex3f(-7 + i, 4, 1);
		glVertex3f(-6.5 + i, 4, 1);
		glVertex3f(-6.5 + i, 4, 7);
		glVertex3f(-7 + i, 4, 7);
	}
	glEnd();
	
	bindPrograms(vertexProg, fragment_prog2);	//board 3
	glNormal3f(0,1,0);
	glBegin(GL_QUADS);
	glTexCoord2f(1.5, -1.2);	glVertex3f(7, -4, 1);
	glTexCoord2f(-2, -1.2);		glVertex3f(1, -4, 1);
	glTexCoord2f(-2, 1.2);		glVertex3f(1, -4, 7);
	glTexCoord2f(1.5, 1.2);		glVertex3f(7, -4, 7);
	glEnd();

    bindPrograms(vertexProg, fragment_prog3);	//board 2
	glEnable (GL_TEXTURE_2D);
   	cgGLSetTextureParameter(image_texture, texture_id);
	cgGLEnableTextureParameter(image_texture);

	glBindTexture( GL_TEXTURE_2D, texture_id );
	glEnable (GL_TEXTURE_2D);
	glNormal3f(0,1,0);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);		glVertex3f(-1, -4, 1);
	glTexCoord2f(0,1);		glVertex3f(-7, -4, 1);
	glTexCoord2f(1,1);		glVertex3f(-7, -4, 7);
	glTexCoord2f(1,0);		glVertex3f(-1, -4, 7);
	glEnd();
	glDisable(GL_TEXTURE_2D);
	
    bindPrograms(vertexProg, fragment_prog1);	//board 1
	glNormal3f(0,1,0);
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);		glVertex3f(1, 4, 1);
	glTexCoord2f(10,0);		glVertex3f(7, 4, 1);
	glTexCoord2f(10,10);	glVertex3f(7, 4, 7);
	glTexCoord2f(0,10);		glVertex3f(1, 4, 7);
	glEnd();

	glutSwapBuffers();
}

static void initPrograms()
{
	cgSetErrorCallback(Handle_Cg_Error);
    context = cgCreateContext();
    assert(cgIsContext(context));
    // Do one-time setup only once; setup Cg programs and textures and set up OpenGL state.
    if (vertexProfile == CG_PROFILE_UNKNOWN || fragmentProfile == CG_PROFILE_UNKNOWN) 
	{
        vertexProfile   = cgGLGetLatestProfile(CG_GL_VERTEX);
        fragmentProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
    }
    cgGLSetOptimalOptions(vertexProfile);
    cgGLSetOptimalOptions(fragmentProfile);

    // Create vertex program.  This is the vertex program used with all of the fragment shaders.
    vertexProg = cgCreateProgramFromFile(context, CG_SOURCE, "../vertex_prog1.cg",vertexProfile, "main", compilerArgs);
    assert(vertexProg);
    cgGLLoadProgram(vertexProg);
	vertexProg2 = cgCreateProgramFromFile(context, CG_SOURCE, "../vertex_prog2.cg",vertexProfile, "main", compilerArgs);
    assert(vertexProg2);
    cgGLLoadProgram(vertexProg2);

    // Turn on manual compilation so the runtime won't try to compile incomplete programs.
    cgSetAutoCompile(context, CG_COMPILE_MANUAL);
	
	fragment_prog1=cgCreateProgramFromFile(context, CG_SOURCE, "../fragment_prog1.cg", fragmentProfile, "main", compilerArgs);
	fragment_prog2=cgCreateProgramFromFile(context, CG_SOURCE, "../fragment_prog2.cg", fragmentProfile, "main", compilerArgs);
	fragment_prog3=cgCreateProgramFromFile(context, CG_SOURCE, "../fragment_prog3.cg", fragmentProfile, "main", compilerArgs);
	fragment_prog4=cgCreateProgramFromFile(context, CG_SOURCE, "../fragment_prog4.cg", fragmentProfile, "main", compilerArgs);
	fragment_prog5=cgCreateProgramFromFile(context, CG_SOURCE, "../fragment_prog5.cg", fragmentProfile, "main", compilerArgs);

    cgCompileProgram(fragment_prog1);
    cgGLLoadProgram(fragment_prog1);
    cgCompileProgram(fragment_prog2);
    cgGLLoadProgram(fragment_prog2);
    cgCompileProgram(fragment_prog3);
    cgGLLoadProgram(fragment_prog3);
    cgCompileProgram(fragment_prog4);
    cgGLLoadProgram(fragment_prog4);
    cgCompileProgram(fragment_prog5);
    cgGLLoadProgram(fragment_prog5);

	//Get the image_tex handle for "fragment_prog3".
	image_texture = cgGetNamedParameter(fragment_prog3, "image_tex");
}

static void Handle_Motion(int x, int y)
{
    if (moveMode) 
	{
		camera_elevate_angle += y-mouseY;
		camera_rotate_angle += x- mouseX;
    }
    mouseX = x;
    mouseY = y;
    glutPostRedisplay();
}

static void Handle_Mouse(int button, int state, int x, int y)
{
	if(button==GLUT_LEFT_BUTTON)
	{
        if (state==GLUT_UP) moveMode = false;
        else 
		{
			moveMode = true;
            mouseX=x;
			mouseY=y;
        }
        glutPostRedisplay();
	}
}

static void Handle_Reshape(int x, int y)
{
    glViewport(0,0,x,y);
    screenWidth = x;
    screenHeight = y;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(25,(float)screenWidth/(float)screenHeight,.1,1000);
    glMatrixMode(GL_MODELVIEW);
}

static void Handle_Keyboard(unsigned char key, int x, int y)
{
	if(key=='q' || key=='Q' || key==27)
	{
        cgDestroyContext(context);  
		exit(0);
    }
}

static void Handle_Cg_Error() 
{
    CGerror err = cgGetError();
    fprintf(stderr, "Cg error: %s\n", cgGetErrorString(err));
    if (err == 1) fprintf(stderr, "last listing:\n%s\n", cgGetLastListing(context));
    fflush(stderr);	
	getchar();
	exit(2);
}
